@@ -11,6 +11,8 @@ index((aggregate_all), 4, aggregate, aggregate).
 index((free_variables), 4, aggregate, aggregate).
 index((ansi_format), 3, ansi_term, ansi_term).
 index((ansi_get_color), 2, ansi_term, ansi_term).
+index((ansi_hyperlink), 2, ansi_term, ansi_term).
+index((ansi_hyperlink), 3, ansi_term, ansi_term).
 index((include), 3, apply, apply).
 index((exclude), 3, apply, apply).
 index((partition), 4, apply, apply).
@@ -287,6 +289,7 @@ index((nodebug), 1, prolog_debug, debug).
 index((debugging), 1, prolog_debug, debug).
 index((debugging), 2, prolog_debug, debug).
 index((list_debug_topics), 0, prolog_debug, debug).
+index((list_debug_topics), 1, prolog_debug, debug).
 index((debug_message_context), 1, prolog_debug, debug).
 index((assertion), 1, prolog_debug, debug).
 index((expects_dialect), 1, prolog_dialect, dialect).
@@ -563,6 +566,7 @@ index((argv_usage), 1, prolog_main, main).
 index((cli_parse_debug_options), 2, prolog_main, main).
 index((cli_enable_development_system), 0, prolog_main, main).
 index((make), 0, make, make).
+index((make_reload_file), 1, make, make).
 index((md5_hash), 3, md5, md5).
 index((new_memory_file), 1, memory_file, memfile).
 index((free_memory_file), 1, memory_file, memfile).
@@ -626,7 +630,6 @@ index((odbc_next_result_set), 1, odbc, odbc).
 index((odbc_close_statement), 1, odbc, odbc).
 index((odbc_clone_statement), 2, odbc, odbc).
 index((odbc_free_statement), 1, odbc, odbc).
-index((odbc_cancel_thread), 1, odbc, odbc).
 index((odbc_current_table), 2, odbc, odbc).
 index((odbc_current_table), 3, odbc, odbc).
 index((odbc_table_column), 3, odbc, odbc).
@@ -704,15 +707,38 @@ index((paxos_quorum_ask), 4, paxos, paxos).
 index((paxos_replicate_key), 3, paxos, paxos).
 index((re_match), 2, pcre, pcre).
 index((re_match), 3, pcre, pcre).
+index((re_matchsub), 3, pcre, pcre).
 index((re_matchsub), 4, pcre, pcre).
 index((re_foldl), 6, pcre, pcre).
 index((re_split), 3, pcre, pcre).
 index((re_split), 4, pcre, pcre).
 index((re_replace), 4, pcre, pcre).
+index((re_replace), 5, pcre, pcre).
 index((re_compile), 3, pcre, pcre).
 index((re_flush), 0, pcre, pcre).
 index((re_config), 1, pcre, pcre).
 index((pdt_install_console), 0, pdt_console, pdt_console).
+index((pengine_create), 1, pengines, pengines).
+index((pengine_ask), 3, pengines, pengines).
+index((pengine_next), 2, pengines, pengines).
+index((pengine_stop), 2, pengines, pengines).
+index((pengine_event), 2, pengines, pengines).
+index((pengine_input), 2, pengines, pengines).
+index((pengine_output), 1, pengines, pengines).
+index((pengine_respond), 3, pengines, pengines).
+index((pengine_debug), 2, pengines, pengines).
+index((pengine_self), 1, pengines, pengines).
+index((pengine_pull_response), 2, pengines, pengines).
+index((pengine_destroy), 1, pengines, pengines).
+index((pengine_destroy), 2, pengines, pengines).
+index((pengine_abort), 1, pengines, pengines).
+index((pengine_application), 1, pengines, pengines).
+index((current_pengine_application), 1, pengines, pengines).
+index((pengine_property), 2, pengines, pengines).
+index((pengine_user), 1, pengines, pengines).
+index((pengine_event_loop), 2, pengines, pengines).
+index((pengine_rpc), 2, pengines, pengines).
+index((pengine_rpc), 3, pengines, pengines).
 index((pengine_writeln), 1, pengines_io, pengines_io).
 index((pengine_nl), 0, pengines_io, pengines_io).
 index((pengine_tab), 1, pengines_io, pengines_io).
@@ -753,6 +779,7 @@ index((run_tests), 1, plunit, plunit).
 index((load_test_files), 1, plunit, plunit).
 index((running_tests), 0, plunit, plunit).
 index((current_test), 5, plunit, plunit).
+index((current_test_unit), 2, plunit, plunit).
 index((test_report), 1, plunit, plunit).
 index((porter_stem), 2, porter_stem, porter_stem).
 index((unaccent_atom), 2, porter_stem, porter_stem).
@@ -791,6 +818,7 @@ index((autoload_all), 0, prolog_autoload, prolog_autoload).
 index((autoload_all), 1, prolog_autoload, prolog_autoload).
 index((set_breakpoint), 4, prolog_breakpoints, prolog_breakpoints).
 index((set_breakpoint), 5, prolog_breakpoints, prolog_breakpoints).
+index((set_breakpoint_condition), 2, prolog_breakpoints, prolog_breakpoints).
 index((delete_breakpoint), 1, prolog_breakpoints, prolog_breakpoints).
 index((breakpoint_property), 2, prolog_breakpoints, prolog_breakpoints).
 index((clause_info), 4, prolog_clause, prolog_clause).
@@ -865,12 +893,15 @@ index((file_name_on_path), 2, prolog_source, prolog_source).
 index((file_alias_path), 2, prolog_source, prolog_source).
 index((path_segments_atom), 2, prolog_source, prolog_source).
 index((directory_source_files), 3, prolog_source, prolog_source).
+index((valid_term_position), 2, prolog_source, prolog_source).
 index((get_prolog_backtrace), 2, prolog_stack, prolog_stack).
 index((get_prolog_backtrace), 3, prolog_stack, prolog_stack).
 index((prolog_stack_frame_property), 2, prolog_stack, prolog_stack).
 index((print_prolog_backtrace), 2, prolog_stack, prolog_stack).
 index((print_prolog_backtrace), 3, prolog_stack, prolog_stack).
 index((backtrace), 1, prolog_stack, prolog_stack).
+index((print_last_choicepoint), 0, prolog_stack, prolog_stack).
+index((print_last_choicepoint), 2, prolog_stack, prolog_stack).
 index((open_prolog_stream), 4, prolog_stream, prolog_stream).
 index((trace), 1, prolog_trace, prolog_trace).
 index((trace), 2, prolog_trace, prolog_trace).
@@ -1177,7 +1208,6 @@ index((current_foreign_library), 2, shlib, shlib).
 index((reload_foreign_libraries), 0, shlib, shlib).
 index((use_foreign_library), 1, shlib, shlib).
 index((use_foreign_library), 2, shlib, shlib).
-index((win_add_dll_directory), 1, shlib, shlib).
 index((snowball), 3, snowball, snowball).
 index((snowball_current_algorithm), 1, snowball, snowball).
 index((tcp_socket), 1, socket, socket).
@@ -1271,6 +1301,7 @@ index((delete_stream_from_pool), 1, stream_pool, streampool).
 index((close_stream_pool), 0, stream_pool, streampool).
 index((dispatch_stream_pool), 1, stream_pool, streampool).
 index((stream_pool_main_loop), 0, stream_pool, streampool).
+index((with_output_to), 3, streams, streams).
 index((dedent_lines), 3, strings, strings).
 index((indent_lines), 3, strings, strings).
 index((indent_lines), 4, strings, strings).
@@ -1324,6 +1355,8 @@ index((get_returns_for_call), 2, tables, tables).
 index((get_residual), 2, tables, tables).
 index((set_pil_on), 0, tables, tables).
 index((set_pil_off), 0, tables, tables).
+index((term_to_json), 3, term_to_json, term_to_json).
+index((term_to_json), 2, term_to_json, term_to_json).
 index((term_hash), 2, terms, terms).
 index((term_hash), 4, terms, terms).
 index((term_size), 2, terms, terms).
